{
  (**
    * JSON serializer and parser
    * Copyright (C) 2002 Rafael R. Sevilla <dido@imperium.ph>
    * This file is part of JSON for Ocaml
    *
    * JSON for OCaml is free software; you can redistribute it and/or
    * modify it under the terms of the GNU Lesser General Public License
    * as published by the Free Software Foundation; either version 2.1 of
    * the License, or (at your option) any later version.
    *
    * JSON for OCaml is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    * Lesser General Public License for more details. 
    *
    * You should have received a copy of the GNU Lesser General Public
    * License along with JSON for OCaml; if not, write to the Free
    * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    * 02111-1307 USA.
    * $Id: json.mll,v 1.2 2003/08/20 09:22:22 didosevilla Exp $
    * @author Rafael R. Sevilla <dido@imperium.ph>
  *)
type token =
    TRUE
  | FALSE
  | INT of int
  | FLOAT of float
  | STRING of string
  | OBRACE
  | CBRACE
  | OBRACK
  | CBRACK
  | COLON
  | COMMA
  | NULL
  | EOF
;;

type jsontype =
    Bool of bool
  | Int of int
  | Float of float
  | String of string
  | Object of (string, jsontype) Hashtbl.t
  | Array of jsontype array
  | Null
;;

exception Eof;;

(** Unescapes the string until the next quote character given by ptr.
  * Returns a tuple consisting of the unescaped string and the offset
  * of the next character after the closing quote.*)
  let rec unescape str ptr =
    let quotechar = String.get str ptr
    in
    let rec scanner quote newstr str ptr =
      if ptr >= String.length str
      then
	raise (Invalid_argument "unescape/scanner") (* unterminated string *)
      else
	let curchar = String.get str ptr
	in
	  if curchar == quote
	  then
	    (newstr, ptr+1)
	  else
	    match curchar with
		'\\' -> (* backslash escapes *)
		  begin
		    match (String.get str (ptr+1)) with
			'b' -> scanner quote (newstr ^ "\b") str (ptr+2)
		      | 't' -> scanner quote (newstr ^ "\t") str (ptr+2)
		      | 'n' -> scanner quote (newstr ^ "\n") str (ptr+2)
		      | 'f' -> scanner quote (newstr ^ "\012") str (ptr+2)
		      | 'r' -> scanner quote (newstr ^ "\r") str (ptr+2)
		      | 'u' ->
			  let chr = Scanf.sscanf (String.sub str (ptr+2) 4)
				      "%x"
				      Char.chr
			  in
			    (scanner quote (newstr ^ Printf.sprintf "%c" chr)
			       str (ptr+6))
		      | _ ->		(* invalid escape sequence *)
			  raise (Invalid_argument "unescape/scanner")
		  end
	      | x -> scanner quote (newstr ^ (Printf.sprintf "%c" x)) str
		  (ptr+1)
    in
      scanner quotechar "" str (ptr+1)
  ;;
}
rule token = parse
    [' ' '\t' '\n'] { token lexbuf }	(* skip blanks *)
  | ['0'-'9']+ { INT(int_of_string(Lexing.lexeme lexbuf)) }
  | "true"    { TRUE }
  | "false"   { FALSE }
  | "null"    { NULL }
  | '{'       { OBRACE }
  | '}'       { CBRACE }
  | '['       { OBRACK }
  | ']'       { CBRACK }
  | ':'       { COLON }
  | ','       { COMMA }
  | ('\"'[^'\"']*'\"') | ('\''[^'\'']*'\'')
	 { STRING(fst (unescape (Lexing.lexeme lexbuf) 0)) }
  | "//" [^'\n'] '\n' { token lexbuf }	(* skip comments *)
  | "/*" _* "*/" { token lexbuf }	(* skip comments *)
  | ['0'-'9']*'.'?['0'-'9']*(('e'|'E')('+'|'-')?['0'-'9']+)?
      {
	  FLOAT(float_of_string(Lexing.lexeme lexbuf))
      }
  | eof       { EOF }
{

let escape_str str =
  let rec escape_str_s str ofs =
    if ofs >= (String.length str)
    then
      ""
    else
      match String.get str ofs with
	  '\\' -> "\\\\" ^ escape_str_s str (ofs+1)
	| '\b' -> "\\b" ^ escape_str_s str (ofs+1)
	| '\t' -> "\\t" ^ escape_str_s str (ofs+1)
	| '\n' -> "\\n" ^ escape_str_s str (ofs+1)
	| '\012' -> "\\f" ^ escape_str_s str (ofs+1)
	| '\r' -> "\\r" ^ escape_str_s str (ofs+1)
	| x -> 
	    begin
	      if (Char.code x) < 20
	      then (Printf.sprintf "%0x" (Char.code x)) ^ escape_str_s
		str (ofs+1)
	      else
		(Printf.sprintf "%c" x) ^ (escape_str_s str (ofs+1))
	    end
  in
    "\"" ^ escape_str_s str 0 ^ "\""
;;

(**
  * This function takes a JSON object and converts it to the standard
  * JSON string representation. *)
let rec serialize typ =
  match typ with
      Bool b1 -> string_of_bool b1
    | Int i -> string_of_int i
    | Float f -> string_of_float f
    | String s -> escape_str s
    | Object o -> (Hashtbl.fold (fun key value rest -> rest ^ escape_str key ^ ":" ^ (serialize value) ^ ",") o "{") ^ "}"
    | Array a -> (Array.fold_left
		    (fun text elem -> text ^ (serialize elem) ^
				     ",") "[" a) ^ "]"
    | Null -> "null"
;;


let saved_token = Stack.create ();;	(* token stack *)

let get_token lexbuf =
  if Stack.is_empty saved_token
  then
    token lexbuf
  else
    Stack.pop saved_token
;;

let save_token tok =
  Stack.push tok saved_token 
;;

(** [deserialize] does all the hard work of deserializing a JSON
  * string into an equivalent JSON type. *)
let deserialize str = 
  let lexbuf = Lexing.from_string str
  in
  let rec deserialize_rec lexbuf =
    let tok = get_token lexbuf
    in
      match tok with
	  TRUE -> (Bool true)
	| FALSE -> (Bool false)
	| INT x -> (Int x)
	| FLOAT x -> (Float x)
	| STRING x -> (String x)
	| OBRACE ->
	    (* decode a JSON object -- actually returns a list of pairs *)
	    let rec object_decode lexbuf =
	      let key = deserialize_rec lexbuf
	      in
	      let colon = get_token lexbuf
	      in
	      let value = deserialize_rec lexbuf
	      in
	      let comma = get_token lexbuf
	      in
	      let next = get_token lexbuf (* slight lookahead *)
	      in
		match key with
		    String k ->
		      begin
			if colon == COLON
			then
			  begin
			    match comma with
				COMMA ->
				  if next == CBRACE
				  then
				    [(k,value)]
				  else
				    begin
				      save_token next ;
				      (k,value)::object_decode lexbuf
				    end
			      | CBRACE -> [(k,value)]
			      | _ -> raise (Invalid_argument "illegal character found")
			  end
			else
			  raise (Invalid_argument "illegal character found")
		      end
		  | _ -> raise (Invalid_argument "key was not a string")
	    in
	    let obj = Hashtbl.create 16
	    in
	      List.iter (fun (key,value) -> Hashtbl.add obj key value)
		(object_decode lexbuf) ; (Object obj)
	| OBRACK ->
	    let rec array_decode lexbuf =
	      let value = deserialize_rec lexbuf
	      in
	      let comma = get_token lexbuf
	      in
	      let next = get_token lexbuf
	      in
		match comma with
		    COMMA ->
		      if next == CBRACK
		      then
			[value]
		      else
			begin
			  save_token next ;
			  value :: array_decode lexbuf
			end
		  | CBRACK -> [value]
		  | _ -> raise (Invalid_argument "illegal character found")
	    in
	      Array (Array.of_list (array_decode lexbuf))
	| NULL -> Null
	| _ -> raise (Invalid_argument "illegal token found")
  in
    deserialize_rec lexbuf
;;
}
